#include "defs.h"
#include "command.h"
#include "gdbcmd.h"
#include "inferior.h"
#include "loongarch-tdep.h"
#include "target.h"
#include "regcache.h"
#include "linux-nat-trad.h"
#include "loongarch-linux-tdep.h"
#include "target-descriptions.h"

#include "gdb_proc_service.h"
#include "gregset.h"

#include "nat/gdb_ptrace.h"
#include <asm/ptrace.h>
#include "inf-ptrace.h"
#include "arch/loongarch-linux-nat.h"
#include "elf/common.h"

class loongarch_linux_nat_target final : public linux_nat_trad_target
{
public:
  void fetch_registers (struct regcache *, int) override;
  void store_registers (struct regcache *, int) override;

  const struct target_desc *read_description () override;

  enum target_xfer_status xfer_partial (enum target_object object,
					const char *annex, gdb_byte *readbuf,
					const gdb_byte *writebuf,
					ULONGEST offset, ULONGEST len,
					ULONGEST *xfered_len) override;

protected:
  CORE_ADDR register_u_offset (struct gdbarch *gdbarch,
			       int regno, int store_p) override;

  void low_new_thread (struct lwp_info *lp) override;

private:
  void loongarch_fetch_regs (struct regcache *regcache, int regno);
  void loongarch_store_regs (struct regcache *regcache, int regno);
};

/* Assume that we have PTRACE_{GET,SET}REGSET et al. support.  If we do not,
   we'll clear this and use PTRACE_PEEKUSR instead.
   问题的关键在于LA64可以跑32位程序和64位程序，如果一个32位GDB调试64位程序，
   ptrace的各参数和返回值也是32位，无法拿到被调试程序的64位寄存器值。所以
   我们尽量不使用PTRACE_{PEEK,POKE}USR。我们对PTRACE_{PEEK,POKE}USR的支持很
   消极，在内核真的出问题或者是简单调试才使用PTRACE_{PEEK,POKE}USR。  */
extern enum tribool have_ptrace_getregset;

static void ensure_ptrace_getregset ()
{
  have_ptrace_getregset = TRIBOOL_TRUE;
}

/* Fetch REGNO (or all registers if REGNO == -1) from the target
   using any working method.  */

void
loongarch_linux_nat_target::fetch_registers (struct regcache *regcache, int regnum)
{
  if (have_ptrace_getregset == TRIBOOL_UNKNOWN)
    ensure_ptrace_getregset ();

  if (sizeof (void *) == 4)
    gdb_assert (have_ptrace_getregset == TRIBOOL_TRUE
		|| register_size (regcache->arch (), gdbarch_tdep (regcache->arch ())->regs.r) != 32);

  if (have_ptrace_getregset == TRIBOOL_TRUE)
    loongarch_fetch_regs (regcache, regnum);
  else
    linux_nat_trad_target::fetch_registers (regcache, regnum);
}

/* Store REGNO (or all registers if REGNO == -1) to the target
   using any working method.  */

void
loongarch_linux_nat_target::store_registers (struct regcache *regcache, int regnum)
{
  if (have_ptrace_getregset == TRIBOOL_UNKNOWN)
    ensure_ptrace_getregset ();

  if (sizeof (void *) == 4)
    gdb_assert (have_ptrace_getregset == TRIBOOL_TRUE
		|| register_size (regcache->arch (), gdbarch_tdep (regcache->arch ())->regs.r) != 32);

  if (have_ptrace_getregset == TRIBOOL_TRUE)
    loongarch_store_regs (regcache, regnum);
  else
    linux_nat_trad_target::store_registers (regcache, regnum);
}

/* linux_nat_target::low_new_thread implementation.  */

void
loongarch_linux_nat_target::low_new_thread (struct lwp_info *lp)
{
  long tid = lp->ptid.lwp ();
}

const struct target_desc *
loongarch_linux_nat_target::read_description ()
{
  int tid = ptid_get_lwp (inferior_ptid);
  if (tid == 0)
    tid = ptid_get_pid (inferior_ptid);
  ensure_ptrace_getregset ();
  return loongarch_linux_read_description_runtime (tid);
}

void
loongarch_linux_nat_target::loongarch_fetch_regs (struct regcache *regcache, int regno)
{
  auto regs = &gdbarch_tdep (regcache->arch ())->regs;
  pid_t tid = get_ptrace_pid (regcache->ptid ());

  do
    {
      if (regs->r < 0)
	break;
      if (regno == -1);
      else if (regs->r <= regno && regno < regs->r + 32);
      else if (regs->pc == regno);
      else
	break;

      loongarch_elf_gregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_PRSTATUS, &iovec));
      loongarch_elf_gregset.supply_regset (NULL, regcache, regno, &regset, sizeof (regset));
    }
  while (0);

  do
    {
      if (regs->f < 0)
	break;
      if (regno == -1);
      else if (regs->f <= regno && regno < regs->f + 32);
      else if (regs->fcc <= regno && regno < regs->fcc + 8);
      else if (regs->fcsr == regno);
      else
	break;

      loongarch_elf_fpregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_FPREGSET, &iovec));
      loongarch_elf_fpregset.supply_regset (NULL, regcache, regno, &regset, sizeof (regset));
    }
  while (0);

  do
    {
      if (regs->scr < 0)
	break;
      else if (regno == -1);
      else if (regs->scr <= regno && regno < regs->scr + 4);
      else if (regs->EFLAG == regno);
      else if (regs->x86_top == regno);
      else
	break;

      loongarch_elf_lbtregset_t regset = {0};
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LBT, &iovec));
      loongarch_elf_lbtregset.supply_regset (NULL, regcache, regno, &regset, sizeof (regset));
    }
  while (0);

  do
    {
      if (regs->vr < 0)
	break;
      if (regno == -1);
      else if (regs->vr <= regno && regno < regs->vr + 32);
      else
	break;

      loongarch_elf_lsxregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LSX, &iovec));
      loongarch_elf_lsxregset.supply_regset (NULL, regcache, regno, &regset, sizeof (regset));
    }
  while (0);

  do
    {
      if (regs->xr < 0)
	break;
      else if (regno == -1);
      else if (regs->xr <= regno && regno < regs->xr + 32);
      else
	break;

      loongarch_elf_lasxregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LASX, &iovec));
      loongarch_elf_lasxregset.supply_regset (NULL, regcache, regno, &regset, sizeof (regset));
    }
  while (0);
}

void
loongarch_linux_nat_target::loongarch_store_regs (struct regcache *regcache, int regno)
{
  auto regs = &gdbarch_tdep (regcache->arch ())->regs;
  pid_t tid = get_ptrace_pid (regcache->ptid ());

  do
    {
      if (regs->r < 0)
	break;
      if (regno == -1);
      else if (regs->r <= regno && regno < regs->r + 32);
      else if (regs->pc == regno);
      else
	break;

      loongarch_elf_gregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_PRSTATUS, &iovec));
      loongarch_elf_gregset.collect_regset (NULL, regcache, regno, &regset, sizeof (regset));
      gdb_assert (0 == ptrace (PTRACE_SETREGSET, tid, NT_PRSTATUS, &iovec));
    }
  while (0);

  do
    {
      if (regs->f < 0)
	break;
      if (regno == -1);
      else if (regs->f <= regno && regno < regs->f + 32);
      else if (regs->fcc <= regno && regno < regs->fcc + 8);
      else if (regs->fcsr == regno);
      else
	break;

      loongarch_elf_fpregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_FPREGSET, &iovec));
      loongarch_elf_fpregset.collect_regset (NULL, regcache, regno, &regset, sizeof (regset));
      gdb_assert (0 == ptrace (PTRACE_SETREGSET, tid, NT_FPREGSET, &iovec));
    }
  while (0);

  do
    {
      if (regs->scr < 0)
	break;
      else if (regno == -1);
      else if (regs->scr <= regno && regno < regs->scr + 4);
      else if (regs->EFLAG == regno);
      else if (regs->x86_top == regno);
      else
	break;

      loongarch_elf_lbtregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LBT, &iovec));
      loongarch_elf_lbtregset.collect_regset (NULL, regcache, regno, &regset, sizeof (regset));
      gdb_assert (0 == ptrace (PTRACE_SETREGSET, tid, NT_LARCH_LBT, &iovec));
    }
  while (0);

  do
    {
      if (regs->vr < 0)
	break;
      if (regno == -1);
      else if (regs->vr <= regno && regno < regs->vr + 32);
      else
	break;

      loongarch_elf_lsxregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LSX, &iovec));
      loongarch_elf_lsxregset.collect_regset (NULL, regcache, regno, &regset, sizeof (regset));
      gdb_assert (0 == ptrace (PTRACE_SETREGSET, tid, NT_LARCH_LSX, &iovec));
    }
  while (0);

  do
    {
      if (regs->xr < 0)
	break;
      else if (regno == -1);
      else if (regs->xr <= regno && regno < regs->xr + 32);
      else
	break;

      loongarch_elf_lasxregset_t regset;
      struct iovec iovec = { .iov_base = &regset, .iov_len = sizeof (regset)};
      gdb_assert (0 == ptrace (PTRACE_GETREGSET, tid, NT_LARCH_LASX, &iovec));
      loongarch_elf_lasxregset.collect_regset (NULL, regcache, regno, &regset, sizeof (regset));
      gdb_assert (0 == ptrace (PTRACE_SETREGSET, tid, NT_LARCH_LASX, &iovec));
    }
  while (0);
}

/* Return the address in the core dump or inferior of register REGNO.  */

CORE_ADDR
loongarch_linux_nat_target::register_u_offset (struct gdbarch *gdbarch,
					       int regno, int store_p)
{
  auto regs = &gdbarch_tdep (gdbarch)->regs;
  /* according to <asm/ptrace.h> */
  if (0 <= regs->r && regs->r <= regno && regno < regs->r + GPR_NUM)
    return GPR_BASE + regno - regs->r;
  else if (regs->pc == regno)
    return PC;
  return -1;
}


/* Implement the to_xfer_partial target_ops method.  */

enum target_xfer_status
loongarch_linux_nat_target::xfer_partial (enum target_object object,
					  const char *annex, gdb_byte *readbuf,
					  const gdb_byte *writebuf,
					  ULONGEST offset, ULONGEST len,
					  ULONGEST *xfered_len)
{
  pid_t pid = ptid_get_pid (inferior_ptid);

  if (object != TARGET_OBJECT_LARCH)
    return linux_nat_trad_target::xfer_partial
	     (object, annex, readbuf, writebuf, offset, len, xfered_len);

  if (strcmp (annex, "cpucfg") == 0)
    {
      if (writebuf)
	return TARGET_XFER_E_IO;
      if (offset % 4 != 0 || offset % 4 != 0)
	return TARGET_XFER_E_IO;
      char t_buf[offset + len];
      struct iovec iovec = { .iov_base = &t_buf, .iov_len = sizeof (t_buf)};
      if (ptrace (PTRACE_GETREGSET, pid, NT_LARCH_CPUCFG, &iovec) < 0)
	{
	  size_t i;
	  for (i = offset / 4; i < (offset + len) / 4 + 1; i++)
	    ((uint32_t *)t_buf)[i] = loongarch_cpucfg (i);
	}
      memcpy (readbuf, t_buf + offset, len);
      *xfered_len = len;
      return TARGET_XFER_OK;
    }

  return TARGET_XFER_E_IO;
}

static loongarch_linux_nat_target the_loongarch_linux_nat_target;

/* Wrapper functions.  These are only used by libthread_db.  */

void
supply_gregset (struct regcache *regcache,
		const gdb_gregset_t/* elf_gregset_t */ *gregset)
{
  loongarch_elf_gregset.supply_regset
    (NULL, regcache, -1, gregset, sizeof (gdb_gregset_t));
}

void
fill_gregset (const struct regcache *regcache,
	      gdb_gregset_t/* elf_gregset_t */ *gregset, int regno)
{
  loongarch_elf_gregset.collect_regset
    (NULL, regcache, regno, gregset, sizeof (gdb_gregset_t));
}

void
supply_fpregset (struct regcache *regcache,
		 const gdb_fpregset_t/* elf_fpregset_t */ *fpregset)
{
  loongarch_elf_fpregset.supply_regset
    (NULL, regcache, -1, fpregset, sizeof (gdb_fpregset_t));
}

void
fill_fpregset (const struct regcache *regcache,
	       gdb_fpregset_t/* elf_fpregset_t */ *fpregset, int regno)
{
  loongarch_elf_fpregset.collect_regset
    (NULL, regcache, regno, fpregset, sizeof (gdb_fpregset_t));
}

void
_initialize_loongarch_linux_nat (void)
{
  add_setshow_boolean_cmd ("show-debug-regs", class_maintenance,
			   &show_debug_regs, _("\
Set whether to show variables that mirror the mips debug registers."), _("\
Show whether to show variables that mirror the mips debug registers."), _("\
Use \"on\" to enable, \"off\" to disable.\n\
If enabled, the debug registers values are shown when GDB inserts\n\
or removes a hardware breakpoint or watchpoint, and when the inferior\n\
triggers a breakpoint or watchpoint."),
			   NULL,
			   NULL,
			   &maintenance_set_cmdlist,
			   &maintenance_show_cmdlist);

  linux_target = &the_loongarch_linux_nat_target;
  add_inf_child_target (&the_loongarch_linux_nat_target);
}
